/*!
\file  gk_types.h
\brief This file contains basic scalar datatype used in GKlib

\date   Started 3/27/2007
\author George
\version\verbatim $Id: gk_types.h 10711 2011-08-31 22:23:04Z karypis $ \endverbatim
*/

#ifndef _GK_TYPES_H_
#define _GK_TYPES_H_

typedef ssize_t         gk_idx_t;         /* index variable */

#endif
